TIMEOUT_IN_SECONDS = 5

# API_MODE
API_MODE_BITALINO = 1
API_MODE_SCIENTISST = 2
API_MODE_JSON = 3

# COM_MODE
COM_MODE_BT = "bt_classic"
COM_MODE_TCP_SERVER = "tcp"
COM_MODE_TCP_AP = "tcp_ap"
COM_MODE_LIST = [COM_MODE_BT, COM_MODE_TCP_SERVER, COM_MODE_TCP_AP]

# CHANNELS
AI1 = 1
AI2 = 2
AI3 = 3
AI4 = 4
AI5 = 5
AI6 = 6
AX1 = 7
AX2 = 8

MAX_BUFFER_SIZE = 4096
